class SchemaVolumeConfiguration:
    SCHEMA_VOLUME_PATH: str = "_storage/schemas"  # path to volume with default schemas


class ProductionSchemaVolumeConfiguration:
    SCHEMA_VOLUME_PATH: str = None
